# Bump this on any log format change. i.e. changing/adding keys or value types
LOG_FORMAT_VERSION = 1

LOG_RECORD_KEY_EVENT = "event"
LOG_RECORD_KEY_HAS_RATING = "has_rating"
LOG_RECORD_KEY_ID = "id"
LOG_RECORD_KEY_INTERNAL_ID = "internal_id"
LOG_RECORD_KEY_METADATA = "metadata"
LOG_RECORD_KEY_PARENT_ID = "parent_id"
LOG_RECORD_KEY_PARENT_TRACE_ID = "parent_trace_id"
LOG_RECORD_KEY_PROJECT = "project"
LOG_RECORD_KEY_PROPERTIES = "properties"
LOG_RECORD_KEY_RATING = "rating"
LOG_RECORD_KEY_SPANS = "spans"
LOG_RECORD_KEY_TIMESTAMP = "timestamp"
LOG_RECORD_KEY_TYPE = "type"
LOG_RECORD_KEY_USER_ID = "user_id"
LOG_RECORD_KEY_VERSION = "version"

LOG_RECORD_PROPERTIES_KEY_START_TIME = "start_time"
LOG_RECORD_PROPERTIES_KEY_END_TIME = "end_time"
LOG_RECORD_PROPERTIES_KEY_LLM_EVENT = "llm_event"
LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT = "diffusion_event"
LOG_RECORD_PROPERTIES_KEY_MODEL = "model"
LOG_RECORD_PROPERTIES_KEY_PROVIDER = "provider"



from typing import Any, Dict, Literal, Optional, Union
from tatara.tatara_types import (
    DiffusionParams,
    DiffusionPrompt,
    LLMPrompt,
    LLMParams,
    LLMUsageMetrics,
)
from .rating import Rating
from .span import Span


class _EmptySpan(Span):
    @property
    def id(self) -> str:
        return ""

    @property
    def trace_id(self) -> str:
        return ""

    @property
    def event(self) -> str:
        return ""

    def end(self, end_time: Optional[float] = None) -> None:
        pass

    def log_llm_success(
        self,
        prompt: LLMPrompt | str,
        output: str,
        params: Optional[LLMParams] = None,
        usage_metrics: Optional[LLMUsageMetrics] = None,
    ) -> None:
        pass

    def log_diffusion_input(
        self,
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ) -> None:
        pass

    def log_diffusion_success_with_image_url(
        self,
        image_url: str,
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ) -> None:
        pass

    def log_diffusion_success_with_image_data(
        self,
        image_data: str,
        image_format: Union[str, Literal["png", "jpg"]],
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ) -> None:
        pass

    def log_rating(self, rating: Rating) -> None:
        pass

    def start_span(self, event: str) -> Span:
        return _EmptySpan()

    def log_metadata(self, metadata: Dict[str, Any]) -> None:
        pass

    def __enter__(self) -> Span:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        pass



import atexit
import json
from abc import ABC, abstractmethod

MIN_FLUSH_EVENTS = 100

class RecorderBase(ABC):
    def __init__(self):
        self.results = []
        # flush_events results on exit
        atexit.register(self.flush_events)

    @abstractmethod
    def flush_events(self):
        pass    
        
    @abstractmethod
    def record_eval_run(self, eval_run):
        pass

class PrintRecorder(RecorderBase):
    def __init__(self):
        super().__init__()

    
    def flush_events(self):
        for eval_run in self.results:
            print(json.dumps(eval_run.to_dict()))
        self.results = []


    def record_eval_run(self, eval_run):
        self.results.append(eval_run)
        self.flush_events()




class FileRecorder(RecorderBase):
    def __init__(self, event_filepath: str):
        super().__init__()
        self.event_filepath = event_filepath

    def record(self, eval_run):
        self.results.append(eval_run)
        if len(self.results) > MIN_FLUSH_EVENTS:
            self.flush_events()

    def flush_events(self):
        with open(self.event_filepath, "a") as f:
            for eval_row in self.results:
                f.write(json.dumps(eval_row.to_dict()) + "\n")
        self.results = []





import requests
from requests.adapters import HTTPAdapter
from typing import Optional, Dict, List, Any
import json
from .endpoints import Endpoints
from requests.models import Response


class TataraNetworkClient:
    def __init__(
        self,
        api_key: Optional[str] = None,
        project: Optional[str] = None,
        is_dev: bool = False,
    ):
        self._headers = {"Authorization": api_key, "Content-Type": "application/json"}
        self.endpoints = Endpoints(is_dev=is_dev)
        self.project = project
        adapter = HTTPAdapter(max_retries=0)
        self._http = requests.Session()
        self._http.mount("https://", adapter)
        self._http.mount("http://", adapter)

    ### LOGS
    def send_logs_post_request(self, logs: str) -> Optional[Response]:
        return self.send_post_request(f"{self.endpoints.log_endpoint}/write", logs)

    ### DATASETS
    def send_create_dataset_post_request(self, dataset_name: str) -> Optional[Response]:
        data = {"name": dataset_name, "project": self.project}
        return self.send_post_request(
            self.endpoints.dataset_endpoint, data=json.dumps(data)
        )

    def send_dataset_get_request(self, dataset_name: str) -> Response:
        return self.send_get_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}"
        )

    def send_insert_records_post_request(
        self, dataset_name: str, records: List[Dict[str, Any]]
    ) -> Optional[Response]:
        data = {"records": records, "project": self.project}
        return self.send_post_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}/insert_records",
            data=json.dumps(data),
        )
    def send_attach_spans_post_request(
        self, dataset_name: str, span_ids: List[str]
    ) -> Optional[Response]:
        data = {"span_ids": span_ids, "project": self.project}

        return self.send_post_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}/attach_spans",
            data=json.dumps(data),
        )

    def send_attach_records_post_request(
        self, dataset_name: str, record_ids: List[str]
    ) -> Optional[Response]:
        data = {"record_ids": record_ids, "project": self.project}

        return self.send_post_request(
            f"{self.endpoints.dataset_endpoint}/{dataset_name}/attach_records",
            data=json.dumps(data),
        )

    ### EVALS
    def send_eval_run_post_request(self, eval_run) -> Optional[Response]:
        eval_run_to_dict = eval_run.to_dict()
        data = {"eval_run": eval_run_to_dict, "project": self.project}
        return self.send_post_request(
            f"{self.endpoints.eval_endpoint}/write", data=json.dumps(data)
        )

    ### GENERIC REQUESTS
    def send_post_request(self, endpoint: str, data: str) -> Optional[Response]:

        response = self._http.request(
            "POST",
            endpoint,
            data=data,
            headers=self._headers,
        )
        response.raise_for_status()
        return response

    def send_get_request(
        self, endpoint: str, params: Optional[dict] = None
    ) -> Response:
        response = self._http.get(
            endpoint,
            headers=self._headers,
            params=params,
        )
        response.raise_for_status()
        return response



from unittest.mock import patch

import pytest

from tatara.tatara import log_span, log_trace


@log_trace(event="sync_event")
def sync_trace_function(x, y):
    return x + y


@log_trace(event="async_event")
async def async_trace_function(x, y):
    return x * y


@patch("tatara.tatara.start_trace")
def test_log_trace_sync(mock_start_trace):
    sync_trace_function(2, 3)
    mock_start_trace.assert_called_once_with(event="sync_event", id_=None, user_id=None)


@patch("tatara.tatara.start_trace")
def test_log_trace_sync_with_id(mock_start_trace):
    sync_trace_function(2, 3, trace_id="test_id")  # type: ignore
    mock_start_trace.assert_called_once_with(
        event="sync_event", id_="test_id", user_id=None
    )


@patch("tatara.tatara.start_trace")
def test_log_trace_sync_with_user_id(mock_start_trace):
    sync_trace_function(2, 3, tatara_user_id="test_user_id")  # type: ignore
    mock_start_trace.assert_called_once_with(
        event="sync_event", id_=None, user_id="test_user_id"
    )


@patch("tatara.tatara.start_trace")
def test_log_trace_sync_with_id_and_user_id(mock_start_trace):
    sync_trace_function(2, 3, trace_id="test_id", tatara_user_id="test_user_id")  # type: ignore
    mock_start_trace.assert_called_once_with(
        event="sync_event", id_="test_id", user_id="test_user_id"
    )


def test_log_trace_fails_without_event():
    with pytest.raises(TypeError):

        @log_trace  # type: ignore
        def f(x, y):
            return x + y

        f(2, 3)


@patch("tatara.tatara.start_trace")
@pytest.mark.asyncio
async def test_log_trace_async(mock_start_trace):
    await async_trace_function(2, 3)
    mock_start_trace.assert_called_once_with(
        event="async_event", id_=None, user_id=None
    )


@log_span(event="sync_event")
def sync_span_function(x, y):
    return x + y


@log_span(event="async_event")
async def async_span_function(x, y):
    return x * y


@patch("tatara.tatara.start_span")
def test_log_span_sync(mock_start_span):
    sync_span_function(2, 3)
    mock_start_span.assert_called_once_with("sync_event", None)


@patch("tatara.tatara.start_span")
def test_log_span_sync_with_parent_event(mock_start_span):
    @log_span(event="sync_event", parent_event="parent_event")
    def f(x, y):
        return x + y

    f(2, 3)
    mock_start_span.assert_called_once_with("sync_event", "parent_event")


def test_log_span_fails_without_event():
    with pytest.raises(TypeError):

        @log_span  # type: ignore
        def f(x, y):
            return x + y

        f(2, 3)


@patch("tatara.tatara.start_span")
def test_log_span_with_caller_event(mock_start_span):
    @log_span  # type: ignore
    def f(x, y):
        return x + y

    f(2, 3, tatara_event="caller_event")  # type: ignore
    mock_start_span.assert_called_once_with("caller_event", None)


@patch("tatara.tatara.start_span")
def test_log_span_with_caller_event_and_parent(mock_start_span):
    @log_span  # type: ignore
    def f(x, y):
        return x + y

    f(2, 3, tatara_event="caller_event", tatara_parent_event="caller_parent_event")  # type: ignore
    mock_start_span.assert_called_once_with("caller_event", "caller_parent_event")


@patch("tatara.tatara.start_span")
@pytest.mark.asyncio
async def test_async(mock_start_span):
    await async_span_function(2, 3)
    mock_start_span.assert_called_once_with("async_event", None)



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List, Optional
from tatara.evals.eval_types import EvalResultWithMetadata, RecordWithEvalResults, EvalResult
from tatara.evals.record import Record
from tatara.evals.recorder import PrintRecorder, RecorderBase
from tatara.evals.tatara_recorder import TataraRecorder
from typing import Any, Dict
from dataclasses import field
from tatara.evals.id_generator import IdGenerator
import typing
import time


@dataclass
class Eval:
    name: str
    eval_fn: Callable[[Record], EvalResult]
    description: Optional[str] = None
    seed: Optional[int] = 42


    def run_no_recorder(self, dataset: Any) -> "EvalRun": # TODO: make this a dataset type
        """
        Run the eval without recording the results anywhere
        """
        single_eval_all_records = []
        for record in dataset.records:
            record_with_single_eval_result = RecordWithEvalResults(
                record_id=record['id'],
                input=record['input'],
                output=record['output'],
                eval_results_with_metadata=[EvalResultWithMetadata(
                    eval_name=self.name,
                    eval_description=self.description if self.description else "",
                    result=self.eval_fn(record),
                )],
            )
            single_eval_all_records.append(record_with_single_eval_result)

        return EvalRun(eval_rows=single_eval_all_records, evals=[self], dataset=dataset)

    def run(self, recorder: RecorderBase, dataset: Any) -> None: # TODO: make this a dataset type
        """
        Run the eval with the RecorderBase
        """
        single_eval_all_records = self.run_no_recorder(dataset=dataset)
        recorder.record_eval_run(single_eval_all_records)
        
    
    def to_dict(self) -> Dict[str, Any]:
        return_type_raw_str = typing.get_type_hints(self.eval_fn).get("return")
        eval_result_type = "unknown"
        if return_type_raw_str == float:
            eval_result_type = "float"
        elif return_type_raw_str == int:
            eval_result_type = "int"
        elif return_type_raw_str == str:
            eval_result_type = "str"
        elif return_type_raw_str == bool:
            eval_result_type = "bool"
        return {
            "name": self.name,
            "description": self.description,
            "eval_result_type": eval_result_type,
        }

@dataclass
class Evals:
    evals: List[Eval]

    def combine_single_eval_results_to_row(
        self, all_eval_runs: List["EvalRun"]
    ) -> List[RecordWithEvalResults]:
        record_with_multiple_eval_results = {}
        for eval in all_eval_runs:
            for record_with_single_eval_result in eval.eval_rows:
                # join the eval result for the record into a single row
                record_id = record_with_single_eval_result.record_id
                if record_id in record_with_multiple_eval_results:
                    # just add the eval result to the existing record in the output
                    record_with_multiple_eval_results[record_id].add(record_with_single_eval_result)
                else:
                    record_with_multiple_eval_results[record_id] = RecordWithEvalResults(
                        record_id=record_id,
                        input=record_with_single_eval_result.input,
                        output=record_with_single_eval_result.output,
                        eval_results_with_metadata=record_with_single_eval_result.eval_results_with_metadata,
                    )
        return list(record_with_multiple_eval_results.values())

    def run(self, dataset: Any, print_only: bool = False) -> None: # TODO: make this a dataset type
        """
        Run evals. If print_only is True, the evals will be printed to the console and not recorded.
        """
        recorder = TataraRecorder() if not print_only else PrintRecorder()
        all_eval_runs = []
        for eval in self.evals:
            eval_run: EvalRun = eval.run_no_recorder(dataset=dataset)
            all_eval_runs.append(eval_run)
        eval_rows = self.combine_single_eval_results_to_row(all_eval_runs)
        eval_run = EvalRun(evals=self.evals, eval_rows=eval_rows, dataset=dataset)
        recorder.record_eval_run(eval_run)



@dataclass
class EvalRun:
    id: str = field(init=False)
    evals: List[Eval]
    eval_rows: List[RecordWithEvalResults]
    dataset: Any # TODO: make this a dataset type
    timestamp: int = field(default_factory=lambda: int(time.time()))
    
    @property
    def num_rows(self) -> int:
        return len(self.eval_rows)

    @property
    def num_evals(self) -> int:
        return sum([eval_row.num_evals for eval_row in self.eval_rows])

    def __post_init__(self):
        self.id = IdGenerator.generate_eval_run_uuid()
        # set the run id on the rows
        for eval_row in self.eval_rows:
            eval_row.eval_run_id = self.id

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "dataset": {"id": self.dataset.id, "name": self.dataset.name} if self.dataset else None,
            "evals": [eval.to_dict() for eval in self.evals],
            "num_rows": self.num_rows,
            "eval_rows": [eval_row.to_dict() for eval_row in self.eval_rows],
            "timestamp": self.timestamp,
        }



from typing import Any, Dict, List, Optional
from .trace import Trace
from .span import Span
from .rating import Rating
from .empty_span import _EmptySpan


class _EmptyTrace(Trace):
    @property
    def id(self) -> str:
        return ""

    @property
    def event(self) -> str:
        return ""

    @property
    def span_ids(self) -> List[str]:
        return []

    def start_span(self, event: str, parent_event: Optional[str]) -> Span:
        return _EmptySpan()

    def end(self) -> None:
        pass

    def log_metadata(self, metadata: Dict[str, Any]) -> None:
        pass

    def log_rating(self, rating: Rating) -> None:
        pass

    def __enter__(self) -> Trace:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass



import atexit
import base64
import json
import queue
import threading
import traceback
from dataclasses import asdict, is_dataclass
from enum import Enum
from typing import Dict, List, Optional

from tatara.network._tatara_network_client import TataraNetworkClient
from ._record_keys import (
    LOG_RECORD_KEY_ID,
    LOG_RECORD_KEY_METADATA,
    LOG_RECORD_KEY_PROPERTIES,
    LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT,
    LOG_RECORD_PROPERTIES_KEY_END_TIME,
    LOG_RECORD_PROPERTIES_KEY_LLM_EVENT,
    LOG_RECORD_PROPERTIES_KEY_START_TIME,
)


def _custom_serializer(obj):
    if isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, bytes):
        return base64.b64encode(obj).decode()
    elif is_dataclass(obj):
        return asdict(obj)
    else:
        return "not serializable"


def _merge_dicts_on_key(output_dict: Dict, d1: Dict, d2: Dict, key: str):
    if key in output_dict:
        output_dict[key] = {**d1[key], **d2[key]}
    else:
        output_dict[key] = d2[key]


# TODO: benchmark this and see if it's actually more performant than a generic recurisve merge
def _merge_logs(log1: Dict, log2: Dict):
    merged_log = dict(log1)
    for key in log2:
        if key == LOG_RECORD_KEY_METADATA:
            _merge_dicts_on_key(merged_log, log1, log2, LOG_RECORD_KEY_METADATA)
        elif key == LOG_RECORD_KEY_PROPERTIES:
            log1_properties = log1[key]
            log2_properties = log2[key]
            if key in merged_log:
                if LOG_RECORD_PROPERTIES_KEY_START_TIME in log2_properties:
                    merged_log[key][
                        LOG_RECORD_PROPERTIES_KEY_START_TIME
                    ] = log2_properties[LOG_RECORD_PROPERTIES_KEY_START_TIME]
                if LOG_RECORD_PROPERTIES_KEY_END_TIME in log2_properties:
                    merged_log[key][
                        LOG_RECORD_PROPERTIES_KEY_END_TIME
                    ] = log2_properties[LOG_RECORD_PROPERTIES_KEY_END_TIME]
                if LOG_RECORD_PROPERTIES_KEY_LLM_EVENT in log2_properties:
                    _merge_dicts_on_key(
                        merged_log[key],
                        log1_properties,
                        log2_properties,
                        LOG_RECORD_PROPERTIES_KEY_LLM_EVENT,
                    )
                if LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT in log2_properties:
                    _merge_dicts_on_key(
                        merged_log[key],
                        log1_properties,
                        log2_properties,
                        LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT,
                    )

            else:
                merged_log[key] = log2[key]
        else:
            merged_log[key] = log2[key]

    return merged_log


# TODO: possibly more efficient to have an indexable queue structure that gets updated on writes
def _merge_logs_and_convert_to_json(logs: List[Dict]) -> List[str]:
    merged_logs = dict()
    for log in logs:
        log_id = log[LOG_RECORD_KEY_ID]
        if log_id in merged_logs:
            merged_logs[log_id] = _merge_logs(merged_logs[log_id], log)

        else:
            merged_logs[log_id] = log

    json_list = [
        json.dumps(log, default=_custom_serializer)
        for log in list(merged_logs.values())
    ]
    return json_list[::-1]


# queue logs and periodically send over the wire
class BackgroundLazyQueueLogger:
    def __init__(
        self, queue_size: int, flush_interval: float, tatara_network_client: TataraNetworkClient,api_key: Optional[str] = None
    ):
        self._flush_lock = threading.RLock()
        self._start_thread_lock = threading.RLock()
        self._log_queue = queue.Queue(queue_size)
        self._flush_interval = flush_interval
        self._thread = None
        self._timer = None
        self._queue_full = threading.Semaphore(value=0)
        self._tatara_network_client = tatara_network_client

        atexit.register(
            lambda: (print("Flushing remaining Tatara logs..."), self._flush())
        )

    @property
    def is_alive(self) -> bool:
        return self._thread is not None and self._thread.is_alive()

    def _start(self) -> None:
        if not self.is_alive:
            with self._start_thread_lock:
                if not self.is_alive:
                    self._thread = threading.Thread(
                        target=self._target,
                        name="tatara.BackgroundLazyQueueLogger",
                        daemon=True,
                    )
                    self._thread.start()

                    self._timer = threading.Timer(
                        interval=self._flush_interval, function=self._flush
                    )
                    self._timer.daemon = True
                    self._timer.start()

    def _target(self):
        while True:
            self._queue_full.acquire()
            try:
                self._flush()
            except Exception:
                traceback.print_exc()

    def _flush(self):
        with self._flush_lock:
            log_items = []
            try:
                for _ in range(self._log_queue.qsize()):
                    log_items.append(self._log_queue.get_nowait())
            except queue.Empty:
                pass

            if len(log_items) == 0:
                return

            json_logs_to_flush = _merge_logs_and_convert_to_json(log_items)

            max_batch_size = 100
            max_request_size = 3 * 1024 * 1024
            while True:
                json_log_str = "["
                batch_size = 0
                while (
                    batch_size < max_batch_size and len(json_log_str) < max_request_size
                ):
                    if len(json_logs_to_flush) > 0:
                        json_log_str += json_logs_to_flush.pop() + ","
                        batch_size += 1
                    else:
                        break

                if json_log_str == "[":
                    break

                # Replace last comma with a closing bracket for valid JSON
                json_log_str = json_log_str[:-1] + "]"
                self._tatara_network_client.send_logs_post_request(json_log_str)

    def log(self, log_dict: Dict):
        self._start()
        try:
            self._log_queue.put_nowait(log_dict)
        except queue.Full:
            self._queue_full.release()
            self._log_queue.put(log_dict)



from dataclasses import dataclass
from typing import Dict, Literal, Optional, Union
from enum import Enum


@dataclass
class LLMUsageMetrics:
    prompt_tokens: int
    completion_tokens: int

    @classmethod
    def from_oai_completion_usage(
        cls, completion_usage
    ):  # openai.types.CompletionUsage
        return cls(
            prompt_tokens=completion_usage.prompt_tokens,
            completion_tokens=completion_usage.completion_tokens,
        )


@dataclass
class LLMPrompt:
    system_prompt: Optional[str] = None
    prompt_template: Optional[str] = None
    input_variables: Optional[Dict[str, str]] = None


@dataclass
class LLMParams:
    frequency_penalty: Optional[float] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    model: Optional[str] = None
    provider: Optional[
        Union[
            str,
            Literal[
                "anthropic",
                "anyscale",
                "fireworks",
                "huggingface",
                "openai",
                "replicate",
                "stabilityai",
                "togetherai",
                "other",
            ],
        ]
    ] = None


@dataclass
class DiffusionPrompt:
    prompt_template: Optional[str] = None
    input_variables: Optional[Dict[str, str]] = None
    negative_prompt: Optional[str] = None


@dataclass
class DiffusionParams:
    steps: Optional[int] = None
    height: Optional[int] = None
    width: Optional[int] = None
    model: Optional[str] = None
    provider: Optional[
        Union[
            str,
            Literal[
                "anthropic",
                "anyscale",
                "fireworks",
                "huggingface",
                "openai",
                "replicate",
                "stabilityai",
                "togetherai",
                "other",
            ],
        ]
    ] = None


class LogType(Enum):
    TRACE = 0
    SPAN = 1
    RATING = 2
    EVAL = 3



from dataclasses import dataclass, field
from typing import Optional

from tatara.evals.model_package import ModelPackage


@dataclass
class Record:
    id: str = field(init=False)
    input: str
    output: str
    model_package: Optional[ModelPackage] = None

    def __post_init__(self):
        # check that input and output are not None
        if not self.input:
            raise ValueError("input cannot be None")
        if not self.output:
            raise ValueError("output cannot be None")

    @property
    def model_input_type(self):
        return self.model_package.model_input_type if self.model_package else None

    @property
    def model_output_type(self):
        return self.model_package.model_output_type if self.model_package else None

    def to_dict(self):
        return {
            "id": self.id,
            "input": self.input,
            "output": self.output,
            "model_package": self.model_package.to_dict()
            if self.model_package
            else None,
        }



from .recorder import RecorderBase, MIN_FLUSH_EVENTS
from tatara.network._tatara_network_client import TataraNetworkClient
from tatara.tatara import _get_network_client

class TataraRecorder(RecorderBase):
    def __init__(self):
        super().__init__()
        self.eval_runs = []
        self.total_rows = 0
        self.tatara_network_client: TataraNetworkClient = _get_network_client()


    def flush_events(self):
        for eval_run in self.eval_runs:
            res = self.tatara_network_client.send_eval_run_post_request(eval_run)
            if res is not None and res.status_code != 200:
                print(f"Failed to send eval run to server: {res.text}")
        self.eval_runs = []
    

    def record_eval_run(self, eval_run):
        if self.total_rows > MIN_FLUSH_EVENTS:
            self.flush_events()
        self.eval_runs.append(eval_run)
        self.total_rows += eval_run.num_rows




import uuid
from typing import List


class IdPrefix:
    RECORD = "r_"
    DATASET = "d_"
    EVAL_RUN = "er_"
    EVAL_RUN_ROW = "err_"


class IdGenerator:
    @staticmethod
    def generate_record_uuid() -> str:
        return IdPrefix.RECORD + str(uuid.uuid4())

    @staticmethod
    def generate_dataset_uuid() -> str:
        return IdPrefix.DATASET + str(uuid.uuid4())

    @staticmethod
    def generate_eval_run_row_uuid() -> str:
        return IdPrefix.EVAL_RUN_ROW + str(uuid.uuid4())

    @staticmethod
    def generate_eval_run_uuid() -> str:
        return IdPrefix.EVAL_RUN + str(uuid.uuid4())

    @staticmethod
    def is_record_id(id: str) -> bool:
        return id.startswith(IdPrefix.RECORD)

    @staticmethod
    def validate_record_ids(ids: List[str]) -> None:
        for id in ids:
            if not IdGenerator.is_record_id(id):
                raise ValueError(f"Invalid record id: {id}")



from abc import ABC, abstractmethod
from typing import Any, Dict, Literal, Optional, Union

from tatara.tatara_types import (
    DiffusionParams,
    DiffusionPrompt,
    LLMPrompt,
    LLMParams,
    LLMUsageMetrics,
)
from .rating import Rating


class Span(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        """ID of the span. Equivalent to <span.event>_<trace_id>"""

    @property
    @abstractmethod
    def trace_id(self) -> str:
        """ID for the trace that this span belongs to"""

    @property
    @abstractmethod
    def event(self) -> str:
        """Caller provided event name of the span"""

    @abstractmethod
    def end(self, end_time: Optional[float] = None) -> None:
        """Ends the span. No other methods may be called on the span after this."""

    @abstractmethod
    def log_llm_success(
        self,
        prompt: LLMPrompt | str,
        output: str,
        params: Optional[LLMParams] = None,
        usage_metrics: Optional[LLMUsageMetrics] = None,
    ) -> None:
        """Log an LLM event. Only one llm/diffusion call allowed per span."""

    @abstractmethod
    def log_diffusion_input(
        self,
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ) -> None:
        """Log a diffusion model input event."""

    @abstractmethod
    def log_diffusion_success_with_image_url(
        self,
        image_url: str,
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ) -> None:
        """Log a diffusion model image creation event. Only one llm/diffusion call allowed per span."""

    @abstractmethod
    def log_diffusion_success_with_image_data(
        self,
        image_data: str,
        image_format: Union[str, Literal["png", "jpeg"]],
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ) -> None:
        """Log a diffusion model image creation event. `image_data` is expected to be a base64-encoded string. Only one llm/diffusion call allowed per span."""

    @abstractmethod
    def log_metadata(self, metadata: Dict[str, Any]) -> None:
        """Log arbitrary metadata to the span"""

    @abstractmethod
    def log_rating(self, rating: Rating) -> None:
        """Log a rating to the span"""

    @abstractmethod
    def start_span(self, event: str) -> "Span":
        """Start a child span with the current span as the parent and the current trace as the root."""

    @abstractmethod
    def __enter__(self) -> "Span":
        """Enter context manager"""

    @abstractmethod
    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """Exit context manager"""



from dataclasses import dataclass, field
from typing import Any, Dict, List, Union

from tatara.evals.id_generator import IdGenerator

RecordId = str

EvalResult = Union[bool, int, float, str]

@dataclass
class EvalResultWithMetadata:
    eval_name: str
    eval_description: str
    result: EvalResult

    def to_dict(self) -> Dict[str, Any]:
        return {
            "eval_name": self.eval_name,
            "eval_description": self.eval_description,
            "result": self.result,
        }


@dataclass
class RecordWithEvalResults:
    id: str = field(init=False)
    record_id: RecordId
    eval_run_id: str = field(init=False) # gets set when we add rows to an eval run
    input: str
    output: str
    eval_results_with_metadata: List[EvalResultWithMetadata]

    def __post_init__(self):
        self.id = IdGenerator.generate_eval_run_row_uuid()

    def add(self, record_with_eval_results: "RecordWithEvalResults") -> None:
        self.eval_results_with_metadata.extend(record_with_eval_results.eval_results_with_metadata)
    
    @property
    def num_evals(self) -> int:
        return len(self.eval_results_with_metadata)

    def __iter__(self):
        return iter(self.eval_results_with_metadata)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "eval_run_id": self.eval_run_id,
            "record_id": self.record_id,
            "input": self.input,
            "output": self.output,
            "eval_results": [{record_with_single_eval.eval_name: record_with_single_eval.result} for record_with_single_eval in self.eval_results_with_metadata],
        }



from inspect import iscoroutinefunction
from typing import Optional, List, Dict, Any, Union, Literal
import wrapt
import logging
from .tatara_logging.trace import Trace
from .tatara_logging.span import Span
from .tatara_logging.empty_span import _EmptySpan
from .tatara_logging.empty_trace import _EmptyTrace
from .tatara_logging.rating import Rating
from .tatara_types import (
    DiffusionPrompt,
    DiffusionParams,
    LogType,
    LLMPrompt,
    LLMParams,
    LLMUsageMetrics,
)
from .tatara_logging.utils import _gen_id_from_trace_and_event
from .tatara_logging._record_keys import (
    LOG_RECORD_KEY_ID,
    LOG_RECORD_KEY_PROJECT,
    LOG_RECORD_KEY_EVENT,
    LOG_RECORD_KEY_INTERNAL_ID,
    LOG_RECORD_KEY_SPANS,
    LOG_RECORD_KEY_USER_ID,
    LOG_FORMAT_VERSION,
    LOG_RECORD_KEY_PARENT_ID,
    LOG_RECORD_KEY_PARENT_TRACE_ID,
    LOG_RECORD_KEY_PROPERTIES,
    LOG_RECORD_KEY_TIMESTAMP,
    LOG_RECORD_KEY_TYPE,
    LOG_RECORD_KEY_VERSION,
    LOG_RECORD_PROPERTIES_KEY_END_TIME,
    LOG_RECORD_PROPERTIES_KEY_START_TIME,
    LOG_RECORD_PROPERTIES_KEY_LLM_EVENT,
    LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT,
    LOG_RECORD_PROPERTIES_KEY_MODEL,
    LOG_RECORD_PROPERTIES_KEY_PROVIDER,
    LOG_RECORD_KEY_METADATA,
)

import time
from .tatara_logging._background_queue_logger import BackgroundLazyQueueLogger
from .tatara_logging.utils import _check_event
import uuid
import os

from .client_state import TataraClientState
from .network._tatara_network_client import TataraNetworkClient


DEFAULT_QUEUE_SIZE = 1000
DEFAULT_FLUSH_INTERVAL = 60.0

_tatara_client_state: Optional[TataraClientState] = None


def init(
    project: str,
    queue_size: int = DEFAULT_QUEUE_SIZE,
    flush_interval: float = DEFAULT_FLUSH_INTERVAL,
    api_key: Optional[str] = None,
    is_dev: Optional[bool] = False,
):
    if api_key is None:
        if os.environ.get("TATARA_API_KEY") is not None:
            api_key = os.environ.get("TATARA_API_KEY")
        else:
            raise ValueError("TATARA_API_KEY environment variable must be set.")

    global _tatara_client_state
    _tatara_client_state = TataraClientState(
        project,
        queue_size=queue_size,
        flush_interval=flush_interval,
        api_key=api_key,
        is_dev=is_dev if is_dev is not None else False,
    )


def current_trace() -> Trace:
    """
    Get the current trace. If there is no current trace, returns an empty trace.
    """
    current_trace = _get_client_state().current_trace.get()

    if current_trace is None:
        logging.log(
            logging.WARN, "No current trace found. Did you forget to start the trace?"
        )
        return _EmptyTrace()

    return current_trace


def current_span() -> Span:
    """
    Get the current trace. If there is no current trace, returns an empty trace.
    """
    current_span = _get_client_state().current_span.get()

    if current_span is None:
        logging.log(
            logging.WARN, "No current span found. Did you forget to start the span?"
        )
        return _EmptySpan()

    return current_span


def rate_trace(trace_id: str, rating: Rating):
    """
    Rate a trace.
    """
    # TODO: trace rating
    pass


def start_trace(
    event: str, id_: Optional[str] = None, user_id: Optional[str] = None
) -> Trace:
    """
    Create a new trace.
    """
    return _TraceImpl(
        event, logger=_get_client_state().bglq_logger, id_=id_, user_id=user_id
    )


def start_span(event: str, parent_event: Optional[str] = None) -> Span:
    """
    Create a new span. Automatically finds the current trace and returns an empty span an active trace is not found.
    """
    current_trace = _get_client_state().current_trace.get()

    if current_trace is None:
        return _EmptySpan()

    return current_trace.start_span(event, parent_event)


def log_diffusion_input(
    trace_id: str,
    event: str,
    prompt: DiffusionPrompt | str,
    params: Optional[DiffusionParams] = None,
):
    log = {
        LOG_RECORD_KEY_ID: _gen_id_from_trace_and_event(trace_id, event),
        LOG_RECORD_KEY_PROJECT: _get_client_state().project,
        LOG_RECORD_KEY_TYPE: LogType.SPAN,
        LOG_RECORD_KEY_TIMESTAMP: time.time(),
        LOG_RECORD_KEY_PROPERTIES: {
            LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT: {
                "prompt": prompt,
                "params": params,
            },
            LOG_RECORD_PROPERTIES_KEY_MODEL: params.model
            if params is not None
            else None,
            LOG_RECORD_PROPERTIES_KEY_PROVIDER: params.provider
            if params is not None
            else None,
        },
    }
    return _get_client_state().bglq_logger.log(log)


def log_diffusion_output_with_image_data(
    trace_id: str,
    event: str,
    image_data: str,
    image_format: Union[str, Literal["png", "jpg"]],
):
    pass


def log_diffusion_output_with_image_url(image_url: str, trace_id: str, event: str):
    log = {
        LOG_RECORD_KEY_ID: _gen_id_from_trace_and_event(trace_id, event),
        LOG_RECORD_KEY_PROJECT: _get_client_state().project,
        LOG_RECORD_KEY_TYPE: LogType.SPAN,
        LOG_RECORD_KEY_TIMESTAMP: time.time(),
        LOG_RECORD_KEY_PROPERTIES: {
            LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT: {
                "image_url": image_url,
            }
        },
    }
    return _get_client_state().bglq_logger.log(log)


def log_diffusion_call_with_image_data(
    image_data: str,
    image_format: Union[str, Literal["png", "jpg"]],
    trace_id: str,
    event: str,
    prompt: DiffusionPrompt | str,
    params: Optional[DiffusionParams] = None,
):
    pass


def log_diffusion_call_with_image_url(
    image_url: str,
    trace_id: str,
    event: str,
    prompt: DiffusionPrompt | str,
    params: Optional[DiffusionParams] = None,
):
    pass


def log_rating_for_span(span_event: str, trace_id: str, rating: Rating):
    """
    Add a rating for a span.
    """
    _get_client_state().log_rating(
        rating=rating, trace_id=trace_id, span_event=span_event
    )


def log_rating_for_trace(id: str, rating: Rating):
    """
    Add a rating for a trace. ID can be either the user-provided id or the trace_id.
    """
    _get_client_state().log_rating(rating=rating, trace_id=id, span_event=None)


def log_trace(event: str):
    if callable(event):
        raise TypeError(
            "The 'log_trace' decorator expects an 'event' argument. Usage: @log_trace(event='your_event')"
        )

    def wrapper(wrapped):
        @wrapt.decorator
        async def _async_trace_wrapper(wrapped, instance, args, kwargs):
            trace_id = kwargs.pop("trace_id", None)
            user_id = kwargs.pop("tatara_user_id", None)

            with start_trace(event=event, id_=trace_id, user_id=user_id):
                return await wrapped(*args, **kwargs)

        @wrapt.decorator
        def _trace_wrapper(wrapped, instance, args, kwargs):
            trace_id = kwargs.pop("trace_id", None)
            user_id = kwargs.pop("tatara_user_id", None)

            with start_trace(event=event, id_=trace_id, user_id=user_id):
                return wrapped(*args, **kwargs)

        if iscoroutinefunction(wrapped):
            return _async_trace_wrapper(wrapped)  # type: ignore
        else:
            return _trace_wrapper(wrapped)  # type: ignore

    return wrapper


def log_span(event: Optional[str] = None, parent_event: Optional[str] = None):
    if callable(event):
        # Decorator was used without arguments, and 'event' is the function to be decorated.
        func = event
        return _create_span_decorator(event=None, parent_event=None)(func)
    else:
        # Decorator was used with string arguments.
        def decorator(func):
            return _create_span_decorator(event=event, parent_event=parent_event)(func)

        return decorator


def _create_span_decorator(event: Optional[str], parent_event: Optional[str] = None):
    def _check_event(event: Optional[str]):
        if event is None:
            raise TypeError(
                """Missing required 'event' argument. The 'log_span' decorator expects either:

1) An 'event' argument directly passed in to the decorator: 

    @log_span(event='your_event')
    def decorated_function():
        # do stuff

2) A 'tatara_event' argument passed from the caller of the decorated function:

    @log_span
    def decorated_function():
        # do stuff
    
    decorated_function(tatara_event='your_event') 
"""
            )

    def wrapper(wrapped):
        @wrapt.decorator
        async def _async_span_wrapper(wrapped, instance, args, kwargs):
            _event = kwargs.pop("tatara_event", event)
            _check_event(_event)
            _parent_event = kwargs.pop("tatara_parent_event", parent_event)
            with start_span(_event, _parent_event) as span:  # noqa: F841
                return await wrapped(*args, **kwargs)

        @wrapt.decorator
        def _span_wrapper(wrapped, instance, args, kwargs):
            _event = kwargs.pop("tatara_event", event)
            _check_event(_event)
            _parent_event = kwargs.pop("tatara_parent_event", parent_event)
            with start_span(_event, _parent_event) as span:  # noqa: F841
                return wrapped(*args, **kwargs)

        if iscoroutinefunction(wrapped):
            return _async_span_wrapper(wrapped)  # type: ignore
        else:
            return _span_wrapper(wrapped)  # type: ignore

    return wrapper


def _get_client_state() -> TataraClientState:
    if _tatara_client_state is None:
        raise Exception(
            "Tatara Client State not initialized. Please call init() before using the client."
        )
    return _tatara_client_state


def _get_network_client() -> TataraNetworkClient:
    if _tatara_client_state is None:
        raise Exception(
            "Tatara Client State not initialized. Please call init() before using the client."
        )
    return _tatara_client_state.tatara_network_client


### IMPLEMENTATION


class _TraceImpl(Trace):
    def __init__(
        self,
        event,
        logger: BackgroundLazyQueueLogger,
        id_: Optional[str] = None,
        user_id: Optional[str] = None,
        start_time: Optional[float] = None,
    ):
        _check_event(event)
        self._event = event
        self._internal_id = "t_" + str(uuid.uuid4())
        self._id = id_ or self._internal_id
        self._user_id = user_id
        self._logger = logger
        now = time.time()
        self._start_time = start_time or now
        self._end_time = None
        self._span_ids = []
        self._metadata = {}
        self._token = _get_client_state().current_trace.set(self)

        self._properties = {
            LOG_RECORD_PROPERTIES_KEY_START_TIME: self._start_time,
            LOG_RECORD_PROPERTIES_KEY_END_TIME: None,
        }

        self._log_record = {
            LOG_RECORD_KEY_PROJECT: _get_client_state().project,
            LOG_RECORD_KEY_TYPE: LogType.TRACE,
            LOG_RECORD_KEY_TIMESTAMP: now,
            LOG_RECORD_KEY_VERSION: LOG_FORMAT_VERSION,
            LOG_RECORD_KEY_INTERNAL_ID: self._internal_id,
            LOG_RECORD_KEY_ID: self._id,
            LOG_RECORD_KEY_USER_ID: self._user_id,
            LOG_RECORD_KEY_EVENT: self._event,
            LOG_RECORD_KEY_SPANS: self._span_ids,
            LOG_RECORD_KEY_PROPERTIES: self._properties,
            LOG_RECORD_KEY_METADATA: self._metadata,
        }

        self._logger.log(self._log_record)

    @property
    def id(self) -> str:
        return self._id

    @property
    def event(self) -> str:
        return self._event

    @property
    def span_ids(self) -> List[str]:
        return self._span_ids

    def _generate_id_for_span_event_and_log(self, event: str) -> str:
        span_id = event + "_" + self.id

        # add suffix if we're calling the same span event multiple times
        suffix = 2
        while span_id in self.span_ids:
            span_id = event + "_" + str(suffix) + "_" + self.id
            suffix += 1

        self._span_ids.append(span_id)

        self._log_record[LOG_RECORD_KEY_SPANS] = self._span_ids
        self._log()

        return span_id

    def start_span(self, event: str, parent_event: Optional[str]):
        self._check_finished()
        _check_event(event)

        span_id = self._generate_id_for_span_event_and_log(event)
        # TODO: handle edge case if multiple spans with the same event are called
        parent_id = (
            self.id
            if (parent_event is None or parent_event == self.event)
            else parent_event + "_" + self.id
        )
        return _SpanImpl(
            event=event,
            id_=span_id,
            parent_id=parent_id,
            trace=self,
            background_queue_logger=_get_client_state().bglq_logger,
            start_time=time.time(),
        )

    def end(self, end_time: Optional[float] = None):
        self._check_finished()

        self._end_time = end_time or time.time()
        _get_client_state().current_trace.reset(self._token)

        self._log_record[LOG_RECORD_KEY_PROPERTIES][
            LOG_RECORD_PROPERTIES_KEY_END_TIME
        ] = self._end_time
        self._log()

    def log_metadata(self, metadata: Dict[str, Any]):
        self._check_finished()

        self._metadata.update(metadata)
        self._log_record[LOG_RECORD_KEY_METADATA] = self._metadata
        self._log()

    # we can log a rating after a trace is finished
    def log_rating(self, rating: Rating):
        return _get_client_state().log_rating(
            rating=rating, trace_id=self.id, span_event=None
        )

    def _check_finished(self):
        if self._end_time:
            raise RuntimeError("Cannot call method on a finished trace.")

    def _log(self):
        self._log_record[LOG_RECORD_KEY_TIMESTAMP] = time.time()
        self._logger.log(self._log_record)

    def __enter__(self) -> Trace:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback

        self.end()


class _SpanImpl(Span):
    def __init__(
        self,
        event: str,
        id_: str,
        parent_id: str,
        trace: "_TraceImpl",
        background_queue_logger: BackgroundLazyQueueLogger,
        start_time: Optional[float] = None,
    ):
        self._event = event
        self._internal_id = "s_" + str(uuid.uuid4())
        self._id = id_
        self._parent_id = parent_id
        self._trace = trace

        # self.parent_span_id = parent_span_id
        self._background_queue_logger = background_queue_logger
        now = time.time()
        self._start_time = start_time or now
        self._end_time = None
        self._properties = {
            LOG_RECORD_PROPERTIES_KEY_START_TIME: self._start_time,
            LOG_RECORD_PROPERTIES_KEY_END_TIME: None,
        }
        self._token = _get_client_state().current_span.set(self)

        self._metadata = {}

        self._log_record = {
            LOG_RECORD_KEY_PROJECT: _get_client_state().project,
            LOG_RECORD_KEY_TYPE: LogType.SPAN,
            LOG_RECORD_KEY_TIMESTAMP: now,
            LOG_RECORD_KEY_VERSION: LOG_FORMAT_VERSION,
            LOG_RECORD_KEY_INTERNAL_ID: self._internal_id,
            LOG_RECORD_KEY_ID: self._id,
            LOG_RECORD_KEY_PARENT_TRACE_ID: self._trace.id,
            LOG_RECORD_KEY_PARENT_ID: self._parent_id,
            LOG_RECORD_KEY_EVENT: self._event,
            LOG_RECORD_KEY_PROPERTIES: self._properties,
            LOG_RECORD_KEY_METADATA: self._metadata,
        }

        self._background_queue_logger.log(self._log_record)

    @property
    def id(self) -> str:
        return self._id

    @property
    def trace_id(self):
        return self._trace.id

    @property
    def event(self):
        return self._event

    def _check_finished(self):
        if self._end_time:
            raise RuntimeError("Cannot call method on a finished trace.")

    def _log(self):
        self._log_record[LOG_RECORD_KEY_TIMESTAMP] = time.time()
        self._background_queue_logger.log(self._log_record)

    def log_llm_success(
        self,
        prompt: str | LLMPrompt,
        output: str,
        params: Optional[LLMParams] = None,
        usage_metrics: Optional[LLMUsageMetrics] = None,
    ):
        self._check_finished()

        llm_event = {
            "prompt": prompt,
            "output": output,
            "params": params,
            "usage_metrics": usage_metrics,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_LLM_EVENT] = llm_event
        self._properties[LOG_RECORD_PROPERTIES_KEY_MODEL] = (
            params.model if params is not None else None
        )
        self._properties[LOG_RECORD_PROPERTIES_KEY_PROVIDER] = (
            params.provider if params is not None else None
        )
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_diffusion_input(
        self,
        prompt: DiffusionPrompt | str,
        params: Optional[DiffusionParams] = None,
    ):
        diffusion_event = {
            "prompt": prompt,
            "params": params,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT] = diffusion_event
        self._properties[LOG_RECORD_PROPERTIES_KEY_MODEL] = (
            params.model if params is not None else None
        )
        self._properties[LOG_RECORD_PROPERTIES_KEY_PROVIDER] = (
            params.provider if params is not None else None
        )
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_diffusion_success_with_image_url(
        self,
        image_url: str,
        prompt: str | DiffusionPrompt,
        params: Optional[DiffusionParams] = None,
        image_url_type: Literal["ephemeral", "permalink", "permalink_copy"] = "ephemeral",
    ):
        self._check_finished()

        diffusion_event = {
            "prompt": prompt,
            "image_url": image_url,
            "params": params,
            "image_url_type": image_url_type,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT] = diffusion_event
        self._properties[LOG_RECORD_PROPERTIES_KEY_MODEL] = (
            params.model if params is not None else None
        )
        self._properties[LOG_RECORD_PROPERTIES_KEY_PROVIDER] = (
            params.provider if params is not None else None
        )
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_diffusion_success_with_image_data(
        self,
        image_data: str,
        image_format: Union[str, Literal["png", "jpg"]],
        prompt: str | DiffusionPrompt,
        params: Optional[DiffusionParams] = None
    ):
        self._check_finished()


        diffusion_event = {
            "image_data": image_data,
            "image_format": image_format,
            "prompt": prompt,
            "params": params,
        }

        self._properties[LOG_RECORD_PROPERTIES_KEY_DIFFUSION_EVENT] = diffusion_event
        self._properties[LOG_RECORD_PROPERTIES_KEY_MODEL] = (
            params.model if params is not None else None
        )
        self._properties[LOG_RECORD_PROPERTIES_KEY_PROVIDER] = (
            params.provider if params is not None else None
        )
        self._log_record[LOG_RECORD_KEY_PROPERTIES] = self._properties
        self._log()

    def log_rating(self, rating: Rating) -> None:
        return _get_client_state().log_rating(
            trace_id=self.trace_id, span_event=self.event, rating=rating
        )

    def log_metadata(self, metadata: Dict[str, Any]):
        self._check_finished()

        self._metadata.update(metadata)
        self._log_record[LOG_RECORD_KEY_METADATA] = self._metadata
        self._log()

    def end(self, end_time: Optional[float] = None) -> None:
        self._check_finished()
        _get_client_state().current_span.reset(self._token)

        self._end_time = end_time or time.time()

        self._log_record[LOG_RECORD_KEY_PROPERTIES][
            LOG_RECORD_PROPERTIES_KEY_END_TIME
        ] = self._end_time
        self._log()

    def start_span(self, event: str) -> Span:
        self._check_finished()
        _check_event(event)

        span_id = self._trace._generate_id_for_span_event_and_log(event)

        return _SpanImpl(
            event=event,
            id_=span_id,
            parent_id=self.id,
            trace=self._trace,
            background_queue_logger=self._background_queue_logger,
            start_time=time.time(),
        )

    def __enter__(self) -> Span:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        del exc_type, exc_value, traceback

        self.end()



from tatara.evals.eval import Eval
from tatara.evals.recorder import PrintRecorder, FileRecorder
from tatara.evals.tatara_recorder import TataraRecorder

__all__ = ["Eval", "TataraRecorder", "PrintRecorder", "FileRecorder"]



from tatara.tatara_logging._background_queue_logger import (
    BackgroundLazyQueueLogger,
)
from unittest.mock import patch
import time
from tatara.network._tatara_network_client import TataraNetworkClient

def _get_logger(max_queue_size=5, flush_interval=60):
    return BackgroundLazyQueueLogger(
        queue_size=max_queue_size, flush_interval=flush_interval,
        tatara_network_client=TataraNetworkClient()
    )


def test_lazy_init():
    logger = _get_logger()
    assert not logger.is_alive


def test_thread_start():
    logger = _get_logger()
    logger._start()
    assert logger.is_alive


def test_log_triggers_start():
    logger = _get_logger()
    logger.log({"message": "test log"})
    assert logger.is_alive


@patch("threading.Thread.start")
def test_log_method(mock_start):
    logger = _get_logger()
    test_log = {"message": "test log"}
    logger.log(test_log)
    assert not logger._log_queue.empty()
    assert logger._log_queue.get() == test_log


@patch(
    "tatara.tatara_logging._background_queue_logger.TataraNetworkClient.send_logs_post_request"
)
def test_sends_logs_when_queue_is_full(mock_send_logs):
    max_queue_size = 5
    logger = _get_logger(max_queue_size=max_queue_size)
    logger._start()

    for i in range(max_queue_size):
        logger.log({"id": f"id_{i}"})

    mock_send_logs.assert_not_called()

    logger.log({"id": "id_last"})

    time.sleep(1)
    mock_send_logs.assert_called()


@patch(
    "tatara.tatara_logging._background_queue_logger.TataraNetworkClient.send_logs_post_request"
)
def test_sends_logs_when_timer_fires(mock_send_logs):
    max_queue_size = 5
    logger = _get_logger(max_queue_size=max_queue_size, flush_interval=1)
    logger._start()

    logger.log({"id": "id_1"})
    mock_send_logs.assert_not_called()

    time.sleep(1.5)
    mock_send_logs.assert_called()


@patch(
    "tatara.tatara_logging._background_queue_logger.TataraNetworkClient.send_logs_post_request"
)
def test_noop_when_timer_fires_on_empty_queue(mock_send_logs):
    logger = _get_logger(flush_interval=1)
    logger._start()
    time.sleep(1.5)

    mock_send_logs.assert_not_called()
